include_guard(GLOBAL)
# enforce cmake do not require CMAK_SYSTEM_ARCH/CMAKE_SYSTEM_PROCESSOR variables
cmake_policy(SET CMP0123 NEW)

list(APPEND CMAKE_MODULE_PATH "${CMAKE_CURRENT_LIST_DIR}")

enable_language(ASM)
set(CMAKE_EXECUTABLE_SUFFIX ".elf")

set(_EDX_ESCAPE "[")
set(_ESC_ERR "${_EDX_ESCAPE}31m")
set(_ESC_INFO "${_EDX_ESCAPE}32m")
set(_ESC_RST "${_EDX_ESCAPE}0m")
# for llvm
if ("${CMAKE_C_COMPILER}" MATCHES ".*/armclang.*\.exe")
	set(ARMCLANG 1)
elseif ("${CMAKE_C_COMPILER}" MATCHES ".*/clang\\..*")
	set(CLANG 1)
elseif ("${CMAKE_C_COMPILER}" MATCHES ".*/.*-gcc\.exe")
	set(GCC 1)
endif()

if (GCC)
	execute_process(
		COMMAND ${CMAKE_C_COMPILER} -v
		OUTPUT_VARIABLE  C_COMPILER_V_OUTPUT
		ERROR_VARIABLE  C_COMPILER_V_OUTPUT
	)
	string(REGEX MATCH "Target:[ \\t]*([^ \t\n]*)" GCC_TARGET "${C_COMPILER_V_OUTPUT}")
	set(GCC_TARGET ${CMAKE_MATCH_1})
	string(REGEX MATCH "([^-]*)-([^-]*)-(.*)" GCC_TARGET_ARCH "${GCC_TARGET}")
	set(GCC_TARGET_ARCH ${CMAKE_MATCH_1})
	string(REPLACE "x86_64" "x64" GCC_TARGET_ARCH ${GCC_TARGET_ARCH})
	set(GCC_TARGET_OS ${CMAKE_MATCH_2})
	set(GCC_TARGET_ABI ${CMAKE_MATCH_3})

	string(REGEX MATCH "gcc version[ \\t]*([0-9\\.]+)" GCC_VERSION "${C_COMPILER_V_OUTPUT}")
	set(GCC_VERSION ${CMAKE_MATCH_1})
	if ("${GCC_VERSION}" VERSION_GREATER_EQUAL "11.0.0")
		SET(EMBEDED_GCC_EXTRA_LD_FLAGS
			-Wl,--no-warn-rwx-segments
		)
	endif()

	add_link_options(
		-mthumb
		-static
		-Wl,--gc-sections
	)

	add_compile_options(
		-mthumb
		-ffunction-sections
		-fdata-sections
		-g
		-gdwarf
	)
elseif(ARMCLANG)
	add_compile_options(--target=arm-arm-none-eabi -Wunused-command-line-argument)
elseif(CLANG)
	set(ENABLE_NEWLIB ON CACHE BOOL "Build with newlib: ON, OFF")
	add_link_options(
		--target=arm-none-eabi
		-Wno-unused-command-line-argument
		-mthumb
		-static
		-Wl,--gc-sections
	)

	add_compile_options(
		--target=arm-none-eabi
		-Wno-unused-command-line-argument
		-mthumb
		-ffunction-sections
		-fdata-sections
		-g
		-gdwarf
	)
endif()

function(select_embeded_spec)
	set(TARGET_SPECS "${ARGV0}")
	if (CLANG)
		if (${ENABLE_NEWLIB})
			add_link_options(--config=newlib.cfg)
			if ("${TARGET_SPECS}" STREQUAL "nano")
				add_link_options(--specs=nano.specs)
				add_link_options(-lnosys --specs=nosys.specs)
			elseif ("${TARGET_SPECS}" STREQUAL "nosys")
				add_link_options(-lnosys --specs=nosys.specs)
			elseif ("${TARGET_SPECS}" STREQUAL "rdimon")
				# should invoke initialise_monitor_handles
				add_link_options(-lrdimon --specs=rdimon.specs)
			elseif ("${TARGET_SPECS}" STREQUAL "rdimon-v2m")
				# should invoke initialise_monitor_handles
				add_link_options(-lrdimon-v2m --specs=rdimon-v2m.specs)
			elseif ("${TARGET_SPECS}" STREQUAL "none")
			else()
				message("${_WARNING_PREFIX} : Unsupported Specs " ${TARGET_SPECS})
			endif()
		else()
			if ("${TARGET_SPECS}" STREQUAL "nano")
				# alias as pico
				add_link_options(-lcrt0-minimal -ldummyhost)
			elseif ("${TARGET_SPECS}" STREQUAL "nosys")
				add_link_options(-lcrt0 -ldummyhost)
			elseif ("${TARGET_SPECS}" STREQUAL "rdimon")
				# should invoke initialise_monitor_handles
				add_link_options(-lcrt0-semihost -lsemihost)
			elseif ("${TARGET_SPECS}" STREQUAL "rdimon-v2m")
				# should invoke initialise_monitor_handles
				add_link_options(-lcrt0-semihost -lsemihost)
			elseif ("${TARGET_SPECS}" STREQUAL "none")
			else()
				message("${_WARNING_PREFIX} : Unsupported Specs " ${TARGET_SPECS})
			endif()
		endif()
	else()
		if ("${TARGET_SPECS}" STREQUAL "nano")
			add_link_options(--specs=nano.specs)
			add_link_options(-lnosys --specs=nosys.specs)
		elseif ("${TARGET_SPECS}" STREQUAL "nosys")
			add_link_options(-lnosys --specs=nosys.specs)
		elseif ("${TARGET_SPECS}" STREQUAL "rdimon")
			# should invoke initialise_monitor_handles
			add_link_options(-lrdimon --specs=rdimon.specs)
		elseif ("${TARGET_SPECS}" STREQUAL "rdimon-v2m")
			# should invoke initialise_monitor_handles
			add_link_options(-lrdimon-v2m --specs=rdimon-v2m.specs)
		elseif ("${TARGET_SPECS}" STREQUAL "none")
		else()
			message("${_WARNING_PREFIX} : Unsupported Specs " ${TARGET_SPECS})
		endif()
	endif()
endfunction()

set(_WARNING_PREFIX "       ${_ESC_ERR}Warning${_ESC_RST}")
if ("$ENV{SDK_TYPE}" STREQUAL "STM32Cube")
	include(${CMAKE_CURRENT_LIST_DIR}/stm32cube.cmake)
elseif("$ENV{SDK_TYPE}" STREQUAL "AirMCU")
	include(${CMAKE_CURRENT_LIST_DIR}/air32.cmake)
elseif("$ENV{SDK_TYPE}" STREQUAL "XHSC")
	include(${CMAKE_CURRENT_LIST_DIR}/xhsc.cmake)
elseif("$ENV{SDK_TYPE}" STREQUAL "GD32")
	include(${CMAKE_CURRENT_LIST_DIR}/gd32.cmake)
elseif("$ENV{SDK_TYPE}" STREQUAL "N32")
	include(${CMAKE_CURRENT_LIST_DIR}/n32.cmake)
elseif("$ENV{SDK_TYPE}" STREQUAL "FM33")
	include(${CMAKE_CURRENT_LIST_DIR}/fm33.cmake)
endif()
